import uuid
import logging
from typing import Optional, Dict, Any, Tuple, Awaitable
from decimal import Decimal

from httpx import HTTPStatusError, HTTPError

from credgem.api.base import BaseAPI
from credgem.api.transactions import TransactionResponse
from credgem.exceptions import InsufficientCreditsError

logger = logging.getLogger(__name__)


class DrawCredits:
    """Context manager for drawing credits from a wallet.
    
    This context manager handles the lifecycle of a credit transaction, including:
    - Creating a hold on credits (optional)
    - Debiting credits
    - Releasing held credits if not debited
    - Handling errors and cleanup
    """
    
    def __init__(
        self,
        client,
        wallet_id: str,
        credit_type_id: str,
        amount: Optional[Decimal] = None,
        description: str = "",
        issuer: str = "",
        transaction_id: Optional[str] = None,
        context: Optional[Dict] = None,
        skip_hold: bool = False
    ):
        """Initialize the DrawCredits context.
        
        Args:
            client: The CredGemClient instance
            wallet_id: The ID of the wallet to draw credits from
            credit_type_id: The type of credits to draw
            amount: The amount of credits to hold/debit (optional if skip_hold=True)
            description: A description of the transaction
            issuer: The issuer of the transaction
            transaction_id: Optional transaction ID for idempotency
            context: Optional context data for the transaction
            skip_hold: Whether to skip the hold step and debit directly
        """
        self.client = client
        self.wallet_id = wallet_id
        self.credit_type_id = credit_type_id
        self.amount = amount
        self.description = description
        self.issuer = issuer
        self.transaction_id = transaction_id
        self.context = context or {}
        self.skip_hold = skip_hold
        
        self.hold_id = None
        self.debit_amount = None
    
    async def _handle_api_call(self, coro: Awaitable[TransactionResponse]) -> Tuple[bool, Optional[TransactionResponse]]:
        """Handle API call with proper error handling."""
        try:
            response = await coro
            print("debug response", response)
            return True, response
        except HTTPError as e:
            print("debug error response", e.response.status_code, e.response.json())
            if e.response.status_code == 409:
                # 409 means the operation was already completed
                # Just return success without the response since we get an error message
                return True, None
            elif e.response.status_code == 422:
                # 422 means insufficient credits
                raise InsufficientCreditsError("Insufficient credits available")
            else:
                raise
    
    async def hold(self):
        """Create a hold on the credits."""
        if self.skip_hold:
            return
        
        if self.hold_id:
            return
        
        if not self.amount:
            raise ValueError("Amount is required for hold operation")
        
        logger.info(
            f"Creating hold for {self.amount} credits for wallet {self.wallet_id}"
        )
        success, response = await self._handle_api_call(
            self.client.hold(
                wallet_id=self.wallet_id,
                amount=self.amount,
                credit_type_id=self.credit_type_id,
                description=self.description,
                issuer=self.issuer,
                transaction_id=self.transaction_id,
                idempotency_key=f"hold_{self.transaction_id}",
                context=self.context
            )
        )
        if success and response:
            self.hold_id = response.id
    
    async def debit(self, amount: Optional[Decimal] = None, context: Optional[Dict] = None):
        """Debit credits from the wallet.
        
        Args:
            amount: Optional amount to debit. If not provided, uses the hold amount.
            context: Optional context to override the initial context.
        """
        if not self.skip_hold and not amount and not self.amount:
            raise ValueError("Amount is required for debit operation")
        
        debit_amount = amount or self.amount
        debit_context = {**self.context, **(context or {})}
        
        logger.info(
            f"Debiting {debit_amount} credits from wallet {self.wallet_id}"
        )
        success, response = await self._handle_api_call(
            self.client.debit(
                wallet_id=self.wallet_id,
                amount=debit_amount,
                credit_type_id=self.credit_type_id,
                description=self.description,
                issuer=self.issuer,
                hold_transaction_id=self.hold_id,
                transaction_id=self.transaction_id,
                idempotency_key=f"debit_{self.transaction_id}",
                context=debit_context
            )
        )
        if success:
            self.debit_amount = debit_amount
    
    async def release(self, context: Optional[Dict] = None):
        """Release held credits.
        
        Args:
            context: Optional context to override the initial context.
        """
        if not self.hold_id:
            return
        
        release_context = {**self.context, **(context or {})}
        
        logger.info(
            f"Releasing hold {self.hold_id} for wallet {self.wallet_id}"
        )
        success, _ = await self._handle_api_call(
            self.client.release(
                wallet_id=self.wallet_id,
                hold_transaction_id=self.hold_id,
                credit_type_id=self.credit_type_id,
                description=self.description,
                issuer=self.issuer,
                transaction_id=self.transaction_id,
                idempotency_key=f"release_{self.transaction_id}",
                context=release_context
            )
        )
        if success:
            self.hold_id = None
    
    async def __aenter__(self):
        """Enter the context and create a hold if needed."""
        await self.hold()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit the context and handle cleanup.
        
        If an exception occurred or debit wasn't called, release the hold.
        """
        if self.hold_id and not self.debit_amount:
            await self.release()
 


from typing import Any, Dict, Optional, Type, TypeVar

from httpx import AsyncClient
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class BaseAPI:
    def __init__(self, client: AsyncClient):
        self._client = client

    async def _get(
        self, path: str, params: Optional[Dict[str, Any]] = None, response_model: Type[T] = None
    ) -> T:
        response = await self._client.get(path, params=params)
        response.raise_for_status()
        data = response.json()
        return response_model.model_validate(data) if response_model else data

    async def _post(
        self, path: str, json: Dict[str, Any], response_model: Type[T] = None
    ) -> T:
        response = await self._client.post(path, json=json)
        response.raise_for_status()
        data = response.json()
        return response_model.model_validate(data) if response_model else data

    async def _put(
        self, path: str, json: Dict[str, Any], response_model: Type[T] = None
    ) -> T:
        response = await self._client.put(path, json=json)
        response.raise_for_status()
        data = response.json()
        return response_model.model_validate(data) if response_model else data

    async def _delete(
        self, path: str, response_model: Type[T] = None
    ) -> T:
        response = await self._client.delete(path)
        response.raise_for_status()
        data = response.json()
        return response_model.model_validate(data) if response_model else data 


class CredGemError(Exception):
    """Base exception for all CredGem SDK errors"""
    pass


class InsufficientCreditsError(CredGemError):
    """Raised when there are insufficient credits for an operation"""
    pass


class InvalidRequestError(CredGemError):
    """Raised when the request is invalid"""
    pass


class AuthenticationError(CredGemError):
    """Raised when there are authentication issues"""
    pass


class APIError(CredGemError):
    """Raised when the API returns an error"""
    pass 


from typing import Dict, List, Optional, Any, Union, Literal
from decimal import Decimal
from datetime import datetime
from enum import Enum
from pydantic import BaseModel, Field

from credgem.api.base import BaseAPI


class TransactionType(str, Enum):
    DEPOSIT = "deposit"
    DEBIT = "debit"
    HOLD = "hold"
    RELEASE = "release"
    ADJUST = "adjust"


class HoldStatus(str, Enum):
    HELD = "held"
    USED = "used"
    RELEASED = "released"
    EXPIRED = "expired"


class TransactionStatus(str, Enum):
    PENDING = "pending"
    COMPLETED = "completed"
    FAILED = "failed"


class BalanceSnapshot(BaseModel):
    available: float
    held: float
    spent: float
    overall_spent: float


class TransactionBase(BaseModel):
    wallet_id: str
    credit_type_id: str
    description: str
    idempotency_key: Optional[str] = Field(default=None, description="Idempotency key")
    issuer: str
    context: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Context for the transaction"
    )


class DepositRequest(TransactionBase):
    type: Literal[TransactionType.DEPOSIT] = Field(default=TransactionType.DEPOSIT)
    amount: Decimal = Field(gt=0, description="Amount to deposit")


class DebitRequest(TransactionBase):
    type: Literal[TransactionType.DEBIT] = Field(default=TransactionType.DEBIT)
    amount: Decimal = Field(gt=0, description="Amount to debit")
    hold_transaction_id: Optional[str] = Field(
        default=None, description="Id of the hold transaction to debit"
    )


class HoldRequest(TransactionBase):
    type: Literal[TransactionType.HOLD] = Field(default=TransactionType.HOLD)
    amount: Decimal = Field(gt=0, description="Amount to hold")


class ReleaseRequest(TransactionBase):
    type: Literal[TransactionType.RELEASE] = Field(default=TransactionType.RELEASE)
    hold_transaction_id: str = Field(description="Id of the hold transaction to release")


class AdjustRequest(TransactionBase):
    type: Literal[TransactionType.ADJUST] = Field(default=TransactionType.ADJUST)
    amount: Decimal = Field(description="Amount to adjust")
    reset_spent: bool = False


class TransactionResponse(BaseModel):
    id: str
    wallet_id: Optional[str] = None
    credit_type_id: str
    description: Optional[str] = None
    issuer: Optional[str] = None
    context: Dict = {}
    status: Optional[str] = None
    hold_status: Optional[str] = None
    payload: Optional[Dict[str, Any]] = None
    balance_snapshot: Optional[Dict[str, float]] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None


class PaginatedTransactionResponse(BaseModel):
    page: int
    page_size: int
    total_count: int
    data: List[TransactionResponse]


class TransactionsAPI(BaseAPI):
    """API client for transaction operations."""
    
    async def hold(
        self,
        wallet_id: str,
        amount: Decimal,
        credit_type_id: str,
        description: str,
        issuer: str,
        transaction_id: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        context: Optional[Dict] = None
    ) -> TransactionResponse:
        """Create a hold on credits in a wallet."""
        data = {
            "type": "hold",
            "credit_type_id": credit_type_id,
            "description": description,
            "issuer": issuer,
            "payload": {
                "amount": float(amount)
            },
            "context": context or {}
        }
        if transaction_id:
            data["id"] = transaction_id
        if idempotency_key:
            data["idempotency_key"] = idempotency_key
        
        return await self._post(
            f"/wallets/{wallet_id}/hold",
            json=data,
            response_model=TransactionResponse
        )
    
    async def debit(
        self,
        wallet_id: str,
        amount: Decimal,
        credit_type_id: str,
        description: str,
        issuer: str,
        hold_transaction_id: Optional[str] = None,
        transaction_id: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        context: Optional[Dict] = None
    ) -> TransactionResponse:
        """Debit credits from a wallet."""
        data = {
            "type": "debit",
            "credit_type_id": credit_type_id,
            "description": description,
            "issuer": issuer,
            "payload": {
                "type": "debit",
                "amount": float(amount)
            },
            "context": context or {}
        }
        if hold_transaction_id:
            data["payload"]["hold_transaction_id"] = hold_transaction_id
        if transaction_id:
            data["id"] = transaction_id
        if idempotency_key:
            data["idempotency_key"] = idempotency_key
        
        return await self._post(
            f"/wallets/{wallet_id}/debit",
            json=data,
            response_model=TransactionResponse
        )
    
    async def release(
        self,
        wallet_id: str,
        hold_transaction_id: str,
        credit_type_id: str,
        description: str,
        issuer: str,
        transaction_id: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        context: Optional[Dict] = None
    ) -> TransactionResponse:
        """Release a hold on credits."""
        data = {
            "type": "release",
            "credit_type_id": credit_type_id,
            "description": description,
            "issuer": issuer,
            "payload": {
                "type": "release",
                "hold_transaction_id": hold_transaction_id
            },
            "context": context or {}
        }
        if transaction_id:
            data["id"] = transaction_id
        if idempotency_key:
            data["idempotency_key"] = idempotency_key
        
        return await self._post(
            f"/wallets/{wallet_id}/release",
            json=data,
            response_model=TransactionResponse
        )
    
    async def deposit(
        self,
        wallet_id: str,
        amount: Decimal,
        credit_type_id: str,
        description: str,
        issuer: str,
        transaction_id: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        context: Optional[Dict] = None
    ) -> TransactionResponse:
        """Deposit credits into a wallet."""
        data = {
            "type": "deposit",
            "credit_type_id": credit_type_id,
            "description": description,
            "issuer": issuer,
            "payload": {
                "type": "deposit",
                "amount": float(amount)
            },
            "context": context or {}
        }
        if transaction_id:
            data["id"] = transaction_id
        if idempotency_key:
            data["idempotency_key"] = idempotency_key
        
        return await self._post(
            f"/wallets/{wallet_id}/deposit",
            json=data,
            response_model=TransactionResponse
        )
    
    async def get(self, transaction_id: str) -> TransactionResponse:
        """Get a transaction by ID"""
        return await self._get(f"/transactions/{transaction_id}", response_model=TransactionResponse)
    
    async def list(
        self,
        wallet_id: Optional[str] = None,
        page: int = 1,
        page_size: int = 50,
    ) -> TransactionResponse:
        """List transactions"""
        params = {"page": page, "page_size": page_size}
        if wallet_id:
            params["wallet_id"] = wallet_id
        return await self._get("/transactions", params=params, response_model=TransactionResponse) 


from typing import Dict, List, Optional
from decimal import Decimal
from datetime import datetime
from enum import Enum
from pydantic import BaseModel

from credgem.api.base import BaseAPI


class TimeGranularity(str, Enum):
    DAY = "day"
    WEEK = "week"
    MONTH = "month"


class WalletActivityPoint(BaseModel):
    timestamp: datetime
    wallet_id: str
    wallet_name: str
    total_transactions: int
    total_deposits: float = 0
    total_debits: float = 0
    total_holds: float = 0
    total_releases: float = 0
    total_adjustments: float = 0


class WalletActivityResponse(BaseModel):
    start_date: datetime
    end_date: datetime
    granularity: TimeGranularity
    points: List[WalletActivityPoint]


class CreditUsagePoint(BaseModel):
    timestamp: datetime
    wallet_id: str
    wallet_name: str
    total_debits: float = 0
    total_holds: float = 0
    total_releases: float = 0
    total_adjustments: float = 0


class CreditTypesUsage(BaseModel):
    credit_type_id: str
    credit_type_name: str
    debit_count: int = 0
    total_amount: float = 0


class CreditUsageResponse(BaseModel):
    credit_type_id: str
    credit_type_name: str
    transaction_count: int
    debits_amount: float = 0


class CreditUsageTimeSeriesPoint(BaseModel):
    timestamp: datetime
    credit_type_id: str
    credit_type_name: str
    transaction_count: int
    debits_amount: float = 0


class CreditUsageTimeSeriesResponse(BaseModel):
    start_date: datetime
    end_date: datetime
    granularity: TimeGranularity
    points: List[CreditUsageTimeSeriesPoint]


class InsightsAPI(BaseAPI):
    async def get_wallet_activity(
        self,
        wallet_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        granularity: TimeGranularity = TimeGranularity.DAY,
    ) -> WalletActivityResponse:
        """Get activity insights for a specific wallet"""
        params = {"granularity": granularity}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()
        return await self._get(
            f"/insights/wallets/{wallet_id}/activity",
            params=params,
            response_model=WalletActivityResponse,
        )

    async def get_credit_usage(
        self,
        wallet_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        granularity: TimeGranularity = TimeGranularity.DAY,
    ) -> CreditUsageTimeSeriesResponse:
        """Get credit usage insights for a specific wallet"""
        params = {"granularity": granularity}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()
        return await self._get(
            f"/insights/wallets/{wallet_id}/credit-usage",
            params=params,
            response_model=CreditUsageTimeSeriesResponse,
        )

    async def get_system_credit_usage(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        granularity: TimeGranularity = TimeGranularity.DAY,
    ) -> CreditUsageTimeSeriesResponse:
        """Get system-wide credit usage insights"""
        params = {"granularity": granularity}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()
        return await self._get(
            "/insights/system/credit-usage",
            params=params,
            response_model=CreditUsageTimeSeriesResponse,
        ) 